"""File to handle all operations around OperationOutcomes"""

import logging

logger: logging.Logger = logging.getLogger("fhirsearchhelper.operationoutcomehelper")


def handle_operation_outcomes(resource: dict) -> None:
    issue_codes = [issue["code"] for issue in resource["issue"] if "code" in issue]
    issue_details = [issue["details"]["text"] for issue in resource["issue"] if "details" in issue and "text" in issue["details"]]
    issue_diagnostics = [issue["diagnostics"] for issue in resource["issue"] if "diagnostics" in issue]
    logger.warning(f"There was an OperationOutcome in the return bundle with codes {issue_codes}")
    logger.warning(f"There was an OperationOutcome in the return bundle with issue details of {issue_details}")
    logger.warning(f"There was an OperationOutcome in the return bundle with issue diagnostics of {issue_diagnostics}")



import pytest
from fhir.resources.R4B.capabilitystatement import CapabilityStatement
from requests import Session
from requests.adapters import HTTPAdapter
from urllib3.util import Retry

from fhirsearchhelper.helpers.capabilitystatement import get_supported_search_params, load_capability_statement
from fhirsearchhelper.models.models import SupportedSearchParams

session: Session = Session()
retries: Retry = Retry(total=5, allowed_methods={"GET", "POST", "PUT", "DELETE"}, status_forcelist=[500])
session.mount("https://", HTTPAdapter(max_retries=retries))
session.mount("http://", HTTPAdapter(max_retries=retries))


def test_load_capability_statement_url() -> None:
    cs: CapabilityStatement = load_capability_statement(session=session, url="https://hapi.fhir.org/baseR4/metadata")

    assert cs.resource_type == "CapabilityStatement"
    assert str(cs.implementation.url) == "https://hapi.fhir.org/baseR4"  # type: ignore


def test_load_capability_statement_file_path() -> None:
    cs: CapabilityStatement = load_capability_statement(session=session, file_path="epic_r4_metadata_edited.json")

    assert cs.resource_type == "CapabilityStatement"
    assert str(cs.implementation.url) == "https://fhir.epic.com/interconnect-fhir-oauth/api/FHIR/R4"  # type: ignore


def test_load_capability_statement_no_args() -> None:
    with pytest.raises(ValueError):
        load_capability_statement(session=session)


def test_load_capability_statement_both_args() -> None:
    cs: CapabilityStatement = load_capability_statement(session=session, url="https://hapi.fhir.org/baseR4/metadata", file_path="epic_r4_metadata_edited.json")

    assert cs.resource_type == "CapabilityStatement"
    assert str(cs.implementation.url) == "https://hapi.fhir.org/baseR4"  # type: ignore


def test_get_supported_search_params_capstate() -> None:
    cs: CapabilityStatement = load_capability_statement(session=session, file_path="epic_r4_metadata_edited.json")

    ssps: list[SupportedSearchParams] = get_supported_search_params(cs=cs)

    assert ssps
    assert isinstance(ssps, list)
    assert all([isinstance(sps, SupportedSearchParams) for sps in ssps])
    assert all([resource.type in [sps.resourceType for sps in ssps] for resource in cs.rest[0].resource if "searchParam" in resource.dict(exclude_none=True)])  # type: ignore



"""File for custom models"""

import logging

from fhir.resources.R4B.capabilitystatement import CapabilityStatementRestResourceSearchParam
from pydantic import BaseModel


class CustomFormatter(logging.Formatter):
    grey: str = "\x1b[38;21m"
    green: str = "\x1b[32m"
    yellow: str = "\x1b[33m"
    red: str = "\x1b[31m"
    bold_red: str = "\x1b[31;1m"
    reset: str = "\x1b[0m"
    format_str: str = "{asctime}   {levelname:8s} --- {name}: {message}"

    FORMATS: dict[int, str] = {
        logging.DEBUG: grey + format_str + reset,
        logging.INFO: green + format_str + reset,
        logging.WARNING: yellow + format_str + reset,
        logging.ERROR: red + format_str + reset,
        logging.CRITICAL: bold_red + format_str + reset,
    }

    def format(self, record) -> str:
        log_fmt: str | None = self.FORMATS.get(record.levelno)
        formatter: logging.Formatter = logging.Formatter(log_fmt, "%m/%d/%Y %I:%M:%S %p", style="{")
        return formatter.format(record)


class SupportedSearchParams(BaseModel):
    resourceType: str
    searchParams: list[CapabilityStatementRestResourceSearchParam]


class QuerySearchParams(BaseModel):
    resourceType: str
    searchParams: dict[str, str]



"""File to handle all operations around Condition Resources"""

import logging
from concurrent.futures import Future, ThreadPoolExecutor, as_completed
from copy import deepcopy
from typing import Any

from fhir.resources.R4B.bundle import Bundle
from fhir.resources.R4B.fhirtypes import BundleEntryType
from requests import Response, Session
from requests.adapters import HTTPAdapter
from urllib3.util import Retry

from .operationoutcomehelper import handle_operation_outcomes

logger: logging.Logger = logging.getLogger("fhirsearchhelper.conditionhelper")

cached_encounter_resources: dict = {}


def expand_single_condition_onset(resource: dict, base_url: str, query_headers: dict):
    if resource["resourceType"] == "OperationOutcome":
        handle_operation_outcomes(resource=resource)
        return None

    session: Session = Session()
    retries: Retry = Retry(total=5, allowed_methods={"GET", "POST", "PUT", "DELETE"}, status_forcelist=[500])
    session.mount("https://", HTTPAdapter(max_retries=retries))
    session.mount("http://", HTTPAdapter(max_retries=retries))

    if any(onset_key in resource for onset_key in ["onsetAge", "onsetDateTime", "onsetPeriod", "onsetRange", "onsetString", "recordedDate"]):
        return resource
    if "encounter" in resource and "reference" in resource["encounter"]:
        encounter_ref: str = resource["encounter"]["reference"]
        if base_url + "/" + encounter_ref in cached_encounter_resources:
            logger.debug("Found Encounter in cached resources")
            encounter_json: dict = cached_encounter_resources[base_url + "/" + encounter_ref]
        else:
            logger.debug(f'Did not find Encounter in cached resources, querying {base_url+"/"+encounter_ref}')
            encounter_lookup: Response = session.get(f"{base_url}/{encounter_ref}", headers=query_headers)
            if encounter_lookup.status_code != 200:
                logger.error(f"The Condition Encounter query responded with a status code of {encounter_lookup.status_code}")
                if encounter_lookup.status_code == 403:
                    logger.error("The 403 code typically means your defined scope does not allow for retrieving this resource. Please check your scope to ensure it includes Encounter.Read.")
                    if "WWW-Authenticate" in encounter_lookup.headers:
                        logger.error(encounter_lookup.headers["WWW-Authenticate"])
                return None
            encounter_json = encounter_lookup.json()
            cached_encounter_resources[base_url + "/" + encounter_ref] = encounter_json
        if "period" in encounter_json and "start" in encounter_json["period"]:
            resource["onsetDateTime"] = encounter_json["period"]["start"]
        else:
            resource["onsetDateTime"] = "9999-12-31"
    else:
        resource["onsetDateTime"] = "9999-12-31"

    return resource


def expand_condition_onset(entry_resource: BundleEntryType) -> dict[str, Any] | None:
    """
    Add condition onset date and time information using an Encounter reference.

    This function is designed to enrich a 'Condition' resource by adding onsetDateTime. If the 'Condition' resource already contains 'onsetDateTime', no changes are made.

    Parameters:
    - entry_resource (dict): A Bundle.entry where resourceType is Condition as a dictionary.

    Returns:
    - dict[str, Any] or None: A Bundle.entry with a modified 'Condition' resource dictionary with the 'onsetDateTime' field added or None if an error occurs during the retrieval of the referenced Encounter.

    If the 'Condition' resource references an Encounter, this function makes an HTTP request to fetch the Encounter resource using the provided 'base_url' and 'query_headers'. If successful,
    it extracts the 'start' field from the Encounter's 'period' and adds it as the 'onsetDateTime' in the 'Condition' resource.

    Errors and Logging:
    - If the Encounter retrieval fails (e.g., due to a non-200 status code), an error message is logged containing the status code and provides information about possible solutions for the error.
    - If a 403 status code is encountered, it suggests that the user's scope may be insufficient and provides guidance on checking the scope to ensure it includes 'Encounter.Read'.
    - If the HTTP response contains 'WWW-Authenticate' headers, they are logged to provide additional diagnostic information.
    """

    entry: dict = entry_resource.dict(exclude_none=True)  # type: ignore
    resource: dict = entry["resource"]
    if resource["resourceType"] == "OperationOutcome":
        handle_operation_outcomes(resource=resource)
        return None

    if any(onset_key in resource for onset_key in ["onsetAge", "onsetDateTime", "onsetPeriod", "onsetRange", "onsetString", "recordedDate"]):
        return entry
    if "encounter" in resource and "reference" in resource["encounter"]:
        encounter_ref: str = resource["encounter"]["reference"]
        if g_base_url + "/" + encounter_ref in cached_encounter_resources:
            logger.debug("Found Encounter in cached resources")
            encounter_json: dict = cached_encounter_resources[g_base_url + "/" + encounter_ref]
        else:
            logger.debug(f'Did not find Encounter in cached resources, querying {g_base_url+"/"+encounter_ref}')
            encounter_lookup: Response = g_session.get(f"{g_base_url}/{encounter_ref}", headers=g_query_headers)
            if encounter_lookup.status_code != 200:
                logger.error(f"The Condition Encounter query responded with a status code of {encounter_lookup.status_code}")
                if encounter_lookup.status_code == 403:
                    logger.error("The 403 code typically means your defined scope does not allow for retrieving this resource. Please check your scope to ensure it includes Encounter.Read.")
                    if "WWW-Authenticate" in encounter_lookup.headers:
                        logger.error(encounter_lookup.headers["WWW-Authenticate"])
                return None
            encounter_json = encounter_lookup.json()
            cached_encounter_resources[g_base_url + "/" + encounter_ref] = encounter_json
        if "period" in encounter_json and "start" in encounter_json["period"]:
            resource["onsetDateTime"] = encounter_json["period"]["start"]
        else:
            resource["onsetDateTime"] = "9999-12-31"
    else:
        resource["onsetDateTime"] = "9999-12-31"

    entry["resource"] = resource
    return entry


def expand_condition_onset_in_bundle(session: Session, input_bundle: Bundle, base_url: str, query_headers: dict = {}) -> Bundle:
    """
    Expand and modify resources within a FHIR Bundle by adding Condition.onsetDateTime using referenced Encounter in Condition.encounter.

    This function takes a FHIR Bundle (`input_bundle`) and iterates through its resources. For each resource of type 'Condition',
    it adds Condition.onset using Condition.encounter.reference.resolve().period.start.

    Parameters:
    - input_bundle (Bundle): The input FHIR Bundle containing resources to be processed.
    - base_url (str): The base URL to be used for resolving references within the resources.
    - query_headers (dict, optional): Additional headers to include in HTTP requests when resolving references, such as a previously received Bearer token in an OAuth 2.0 workflow (default: {}).

    Returns:
    - Bundle: A modified FHIR Bundle with resources expanded to include Condition.onsetDateTime, or the original input Bundle if any errors occurred when trying to GET the Encounters.
    """
    global cached_encounter_resources, g_session, g_base_url, g_query_headers
    g_session, g_base_url, g_query_headers = session, base_url, query_headers

    returned_resources: list[BundleEntryType] = input_bundle.entry
    output_bundle: dict = deepcopy(input_bundle).dict(exclude_none=True)

    with ThreadPoolExecutor() as executor:
        future_to_entry: dict[Future[dict[str, Any] | None], BundleEntryType] = {executor.submit(expand_condition_onset, entry): entry for entry in returned_resources}

        expanded_entries: list[dict[str, Any] | None] = []
        for future in as_completed(future_to_entry):
            entry: dict[str, Any] | None = future.result()
            if entry:
                expanded_entries.append(entry)

    expanded_entries_clean: list[dict[str, Any]] = [entry for entry in expanded_entries if entry]

    if len(cached_encounter_resources.keys()) != 0:
        cached_encounter_resources = {}

    output_bundle["entry"] = expanded_entries_clean
    return Bundle.parse_obj(output_bundle)



"""File to handle all operations around Medication-related Resources"""

import logging
from concurrent.futures import Future, ThreadPoolExecutor, as_completed
from copy import deepcopy
from typing import Any

from fhir.resources.R4B.bundle import Bundle
from fhir.resources.R4B.fhirtypes import BundleEntryType
from requests import Response, Session
from requests.adapters import HTTPAdapter
from urllib3.util import Retry

from .operationoutcomehelper import handle_operation_outcomes

logger: logging.Logger = logging.getLogger("fhirsearchhelper.medicationhelper")

cached_medication_resources: dict = {}


def expand_single_medication_reference(resource: dict, base_url: str, query_headers: dict):
    if resource["resourceType"] == "OperationOutcome":
        handle_operation_outcomes(resource=resource)
        return resource

    session: Session = Session()
    retries: Retry = Retry(total=5, allowed_methods={"GET", "POST", "PUT", "DELETE"}, status_forcelist=[500])
    session.mount("https://", HTTPAdapter(max_retries=retries))
    session.mount("http://", HTTPAdapter(max_retries=retries))

    if "medicationReference" in resource:
        med_ref: str = resource["medicationReference"]["reference"]
        if base_url + "/" + med_ref in cached_medication_resources:
            logger.debug("Found Medication in cached resources")
            med_code_concept: dict[str, list[dict]] = cached_medication_resources[base_url + "/" + med_ref]["code"]
        else:
            logger.debug(f'Did not find Medication in cached resources, querying {base_url+"/"+med_ref}')
            med_lookup: Response = session.get(f"{base_url}/{med_ref}", headers=query_headers)
            if med_lookup.status_code != 200:
                logger.error(f"The MedicationRequest Medication query responded with a status code of {med_lookup.status_code}")
                if med_lookup.status_code == 403:
                    logger.error("The 403 code typically means your defined scope does not allow for retrieving this resource. Please check your scope to ensure it includes Medication.Read.")
                    if "WWW-Authenticate" in med_lookup.headers:
                        logger.error(med_lookup.headers["WWW-Authenticate"])
                return None
            cached_medication_resources[base_url + "/" + med_ref] = med_lookup.json()
            med_code_concept = med_lookup.json()["code"]
        resource["medicationCodeableConcept"] = med_code_concept
        del resource["medicationReference"]

    return resource


def expand_medication_reference(entry_resource: BundleEntryType) -> dict[str, Any] | None:
    """
    Expand a MedicationReference within a Bundle.entry.resource.MedicationRequest resource to a MedicationCodeableConcept.

    This function takes a Bundle.entry where the entry.resource is of type MedicationRequest, and if it contains a MedicationReference, it expands it into a MedicationCodeableConcept by resolving the reference using an HTTP request.

    Parameters:
    - entry_resource (dict): A Bundle.entry where resourceType is MedicationRequest as a dictionary.

    Returns:
    - dict: The expanded MedicationRequest resource with MedicationCodeableConcept.

    Errors and Logging:
    - If the Medication retrieval fails (e.g., due to a non-200 status code), an error message is logged containing the status code and provides information about possible solutions for the error.
    - If a 403 status code is encountered, it suggests that the user's scope may be insufficient and provides guidance on checking the scope to ensure it includes 'Medication.Read'.
    - If the HTTP response contains 'WWW-Authenticate' headers, they are logged to provide additional diagnostic information.

    """

    entry: dict[str, Any] = entry_resource.dict(exclude_none=True)  # type: ignore
    resource: dict[str, Any] = entry["resource"]
    if resource["resourceType"] == "OperationOutcome":
        handle_operation_outcomes(resource=resource)
        return resource

    if "medicationReference" in resource:
        med_ref: str = resource["medicationReference"]["reference"]
        if g_base_url + "/" + med_ref in cached_medication_resources:
            logger.debug("Found Medication in cached resources")
            med_code_concept: dict[str, list[dict]] = cached_medication_resources[g_base_url + "/" + med_ref]["code"]
        else:
            logger.debug(f'Did not find Medication in cached resources, querying {g_base_url+"/"+med_ref}')
            med_lookup: Response = g_session.get(f"{g_base_url}/{med_ref}", headers=g_query_headers)
            if med_lookup.status_code != 200:
                logger.error(f"The MedicationRequest Medication query responded with a status code of {med_lookup.status_code}")
                if med_lookup.status_code == 403:
                    logger.error("The 403 code typically means your defined scope does not allow for retrieving this resource. Please check your scope to ensure it includes Medication.Read.")
                    if "WWW-Authenticate" in med_lookup.headers:
                        logger.error(med_lookup.headers["WWW-Authenticate"])
                return None
            cached_medication_resources[g_base_url + "/" + med_ref] = med_lookup.json()
            med_code_concept = med_lookup.json()["code"]
        resource["medicationCodeableConcept"] = med_code_concept
        del resource["medicationReference"]
        entry["resource"] = resource

    return entry


def expand_medication_references_in_bundle(session: Session, input_bundle: Bundle, base_url: str, query_headers: dict = {}) -> Bundle:
    """
    Expand MedicationReferences into MedicationCodeableConcepts for all MedicationRequest entries in a Bundle.

    This function takes a FHIR Bundle containing MedicationRequest resources and expands MedicationReferences into MedicationCodeableConcepts for each entry.

    Parameters:
    - input_bundle (Bundle): The input FHIR Bundle containing MedicationRequest resources to be processed.
    - base_url (str): The base URL used for making HTTP requests to resolve MedicationReferences.
    - query_headers (dict, optional): Additional headers for the HTTP requests (default: {}).

    Returns:
    - Bundle: A modified FHIR Bundle with expanded MedicationReferences or the input Bundle if an error ocurred during expansion.

    The function creates a new Bundle, leaving the original input Bundle unchanged.
    """

    global cached_medication_resources, g_session, g_base_url, g_query_headers
    g_session, g_base_url, g_query_headers = session, base_url, query_headers

    returned_resources: list[BundleEntryType] = input_bundle.entry
    output_bundle: dict = deepcopy(input_bundle).dict(exclude_none=True)

    with ThreadPoolExecutor() as executor:
        future_to_entry: dict[Future[dict[str, Any] | None], BundleEntryType] = {executor.submit(expand_medication_reference, entry): entry for entry in returned_resources}

        expanded_entries: list[dict[str, Any] | None] = []
        for future in as_completed(future_to_entry):
            entry: dict[str, Any] | None = future.result()
            if entry:
                expanded_entries.append(entry)

    expanded_entries_clean: list[dict[str, Any]] = [entry for entry in expanded_entries if entry]

    if len(cached_medication_resources.keys()) != 0:
        cached_medication_resources = {}

    output_bundle["entry"] = expanded_entries_clean
    return Bundle.parse_obj(output_bundle)



from fhir.resources.R4B.bundle import Bundle
from requests import Session
from requests.adapters import HTTPAdapter
from urllib3.util import Retry

from fhirsearchhelper.helpers.documenthelper import expand_document_references_in_bundle

session: Session = Session()
retries: Retry = Retry(total=5, allowed_methods={"GET", "POST", "PUT", "DELETE"}, status_forcelist=[500])
session.mount("https://", HTTPAdapter(max_retries=retries))
session.mount("http://", HTTPAdapter(max_retries=retries))


def test_expand_document_references_no_binary() -> None:
    dc_bundle = Bundle.parse_file("./tests/resources/DocumentReferencesNoBinary.json")

    output = expand_document_references_in_bundle(session=session, input_bundle=dc_bundle, base_url="https://hapi.fhir.org/baseR4/")

    assert isinstance(output, Bundle)
    assert output == dc_bundle



"""File to perform filtering of returned FHIR resources using output from gap analysis"""

import json
import logging
from copy import deepcopy
from pathlib import Path

import fhirpathpy
from fhir.resources.R4B.bundle import Bundle
from fhir.resources.R4B.fhirtypes import BundleEntryType

from ..models.models import QuerySearchParams

logger: logging.Logger = logging.getLogger("fhirsearchhelper.fhirfilter")


def filter_bundle(input_bundle: Bundle, search_params: QuerySearchParams, gap_analysis_output: list[str]) -> Bundle:
    """Function that takes an input bundle, the original search params, and the output from the gap analysis to filter a Bundle"""

    logger.debug("Filtering Bundle using gap analysis output...")

    if not gap_analysis_output:
        return input_bundle

    returned_resources: list[BundleEntryType] = input_bundle.entry
    filtered_entries: list[BundleEntryType] = []
    output_bundle: Bundle = deepcopy(input_bundle)

    for filter_sp in gap_analysis_output:
        filter_sp_value: str | int = search_params.searchParams[filter_sp]
        if "-" in filter_sp:
            filter_sp = filter_sp[0].lower() + "".join(x.capitalize() for x in filter_sp.lower().split("-"))[1:]
        logger.debug(f"Working on filtering for search parameter {filter_sp}")
        match filter_sp:
            case "code":
                code_sp_multiple_codes = filter_sp_value.split("%2C")
                if len(code_sp_multiple_codes) > 1:
                    logger.debug(f"There are a total of {len(code_sp_multiple_codes)} codes in the search parameter")
                for code_sp_single_code in code_sp_multiple_codes:
                    code_sp_split: list[str] = code_sp_single_code.split("%7C")
                    if len(code_sp_split) == 2:  # Case when there is a | separator
                        code_sp_system: str = code_sp_split[0]
                        code_sp_code: str = code_sp_split[1]
                    else:
                        code_sp_system = ""
                        code_sp_code = code_sp_split[0]
                    for entry in returned_resources:
                        if entry.resource.resource_type == "MedicationRequest":  # type: ignore
                            if "coding" not in entry.resource.medicationCodeableConcept.dict():  # type: ignore
                                logger.debug("Code does not have a coding, this MedicationRequest resource does not match")
                                continue
                            if code_sp_system and list(filter(lambda x: x.system == code_sp_system and x.code == code_sp_code, entry.resource.medicationCodeableConcept.coding)):  # type: ignore
                                logger.debug("Found MedicationRequest that matched both system and code for code element")
                                filtered_list = [
                                    (idx, med)
                                    for idx, med in enumerate(entry.resource.medicationCodeableConcept.coding)  # type: ignore
                                    if med.system == code_sp_system and med.code == code_sp_code
                                ]
                                swap_entries = entry.resource.medicationCodeableConcept.coding[filtered_list[0][0]], entry.resource.medicationCodeableConcept.coding[0]  # type: ignore
                                entry.resource.medicationCodeableConcept.coding[0], entry.resource.medicationCodeableConcept.coding[filtered_list[0][0]] = swap_entries  # type: ignore
                                filtered_entries.append(entry)
                            elif any([coding.code == code_sp_code for coding in entry.resource.medicationCodeableConcept.coding]):  # type: ignore
                                logger.debug("Found MedicationRequest that matches code (system was not provided in original query)")
                                filtered_list = [(idx, med) for idx, med in enumerate(entry.resource.medicationCodeableConcept.coding) if med.code == code_sp_code]  # type: ignore
                                swap_entries = entry.resource.medicationCodeableConcept.coding[filtered_list[0][0]], entry.resource.medicationCodeableConcept.coding[0]  # type: ignore
                                entry.resource.medicationCodeableConcept.coding[0], entry.resource.medicationCodeableConcept.coding[filtered_list[0][0]] = swap_entries  # type: ignore
                                filtered_entries.append(entry)
                        else:
                            if "coding" not in entry.resource.code.dict():  # type: ignore
                                logger.debug(f"Code does not have a coding, this {entry.resource.resource_type} resource does not match")  # type: ignore
                                continue
                            if code_sp_system and list(filter(lambda x: x.system == code_sp_system and x.code == code_sp_code, entry.resource.code.coding)):  # type: ignore
                                logger.debug("Found resource that matched both system and code for code element")
                                filtered_entries.append(entry)
                            elif any([coding.code == code_sp_code for coding in entry.resource.code.coding]):  # type: ignore
                                logger.debug("Found resource that matches code (system was not provided in original query)")
                                filtered_entries.append(entry)
            case "category":
                category_sp_split: list[str] = filter_sp_value.split("|")
                if len(category_sp_split) == 2:  # Case when there is a | separator
                    category_sp_system: str = category_sp_split[0]
                    category_sp_code: str = category_sp_split[1]
                else:
                    category_sp_system = ""
                    category_sp_code = category_sp_split[0]
                for entry in returned_resources:
                    if category_sp_system and list(filter(lambda x: x.system == category_sp_system and x.code == category_sp_code, entry.resource.category)):  # type: ignore
                        filtered_entries.append(entry)
                    elif any([coding.code == category_sp_code for coding in entry.resource.category.coding]):  # type: ignore
                        filtered_entries.append(entry)
            case "clinicalStatus":
                for entry in returned_resources:
                    if entry.dict(exclude_none=True)["resource"][filter_sp]["coding"][0]["code"] in filter_sp_value.split(","):  # type: ignore
                        filtered_entries.append(entry)
            case _:
                for entry in returned_resources:
                    if entry.dict(exclude_none=True)["resource"][filter_sp] == filter_sp_value:  # type: ignore
                        filtered_entries.append(entry)

    output_bundle.entry = filtered_entries
    output_bundle.total = len(filtered_entries)  # type: ignore

    return output_bundle


def filter_bundle_new(input_bundle: Bundle, search_params: QuerySearchParams, gap_analysis_output: list[str]) -> Bundle:
    """Function that takes an input bundle, the original search params, and the output from the gap analysis to filter a Bundle"""

    logger.debug("Filtering Bundle using gap analysis output...")

    if not gap_analysis_output:
        return input_bundle

    returned_resources: list[BundleEntryType] = input_bundle.entry
    filtered_entries: list[BundleEntryType] = []
    output_bundle: Bundle = deepcopy(input_bundle)

    with open(f"{Path(__file__).parents[1]}/resources/fhir_r4_search_params.json", "r") as fin:
        r4_sp_json = json.load(fin)

    for filter_sp in gap_analysis_output:
        filter_sp_value: str | int | dict = search_params.searchParams[filter_sp]
        print("Filter sp value", filter_sp_value)
        filter_sp_fhirpath = r4_sp_json[search_params.resourceType][filter_sp]["fhirpath"]

        for entry in returned_resources:
            resource_sp_output = fhirpathpy.evaluate(entry.resource, filter_sp_fhirpath)  # type: ignore
            if isinstance(filter_sp_value, str) and "%7C" in filter_sp_value:
                logger.debug("The filter search parameter value is of type code")
                filter_sp_value = {"system": filter_sp_value.split("%7C")[0], "code": filter_sp_value.split("%7C")[1]}
                logger.debug("Updated filter search parameter value to be a dictionary")
                print(filter_sp_value)
            if filter_sp_value in resource_sp_output:
                filtered_entries.append(entry)

    output_bundle.entry = filtered_entries
    output_bundle.total = len(filtered_entries)  # type: ignore

    return output_bundle



"""File to handle all operations around a CapabilityStatement"""

import logging
import os
from pathlib import Path

from fhir.resources.R4B.capabilitystatement import CapabilityStatement
from requests import Session

from ..models.models import SupportedSearchParams

logger: logging.Logger = logging.getLogger("fhirsearchhelper.capabilitystatement")


def load_capability_statement(session: Session, url: str | None = None, file_path: str | None = None) -> CapabilityStatement:
    """Function to load a CapabilityStatement into memory"""

    if url and file_path:
        logger.info("Defaulting to url...")

    if url:
        try:
            cap_statement: dict = session.get(url, headers={"Accept": "application/json"}).json()
        except Exception as exc:
            logger.error("Something went wrong trying to access the CapabilityStatement via URL")
            raise exc

        try:
            cap_statement_object: CapabilityStatement = CapabilityStatement.parse_obj(cap_statement)
        except Exception as exc:
            logger.error("Something went wrong when trying to turn the retrieved cap statement into a CapabilityStatement object")
            raise exc
    elif file_path:
        if os.path.isfile(f"{Path(__file__).parents[1]}/capabilitystatements/{file_path}"):
            logger.info(f"Found file {file_path} in the CapabilityStatements folder")
            file_path = f"{Path(__file__).parents[1]}/capabilitystatements/{file_path}"
        cap_statement_object = CapabilityStatement.parse_file(file_path)
    else:
        raise ValueError("You need to pass a url, an option to specify a preloaded CapabilityStatement, or a file path.")

    return cap_statement_object


def get_supported_search_params(cs: CapabilityStatement) -> list[SupportedSearchParams]:
    """Function to pull out supported search parameters from a capability statement"""

    return [SupportedSearchParams(resourceType=resource.type, searchParams=resource.searchParam) for resource in cs.rest[0].resource if resource.searchParam]  # type: ignore



"""File to handle all operations around Medication-related Resources"""

import base64
import logging
from concurrent.futures import Future, ThreadPoolExecutor, as_completed
from copy import deepcopy
from typing import Any

import html2text
from fhir.resources.R4B.bundle import Bundle
from fhir.resources.R4B.fhirtypes import BundleEntryType
from requests import Response, Session
from requests.adapters import HTTPAdapter
from requests.exceptions import JSONDecodeError
from urllib3.util import Retry

from .operationoutcomehelper import handle_operation_outcomes

logger: logging.Logger = logging.getLogger("fhirsearchhelper.documenthelper")

cached_binary_resources: dict = {}


def expand_single_document_reference_content(resource: dict, base_url: str, query_headers: dict):
    if resource["resourceType"] == "OperationOutcome":
        handle_operation_outcomes(resource=resource)
        return resource

    session: Session = Session()
    retries: Retry = Retry(total=5, allowed_methods={"GET", "POST", "PUT", "DELETE"}, status_forcelist=[500])
    session.mount("https://", HTTPAdapter(max_retries=retries))
    session.mount("http://", HTTPAdapter(max_retries=retries))

    for i, content in enumerate(resource["content"]):
        if "url" in content["attachment"]:
            binary_url: str = content["attachment"]["url"]
            if base_url + "/" + binary_url in cached_binary_resources:
                logger.debug("Found Binary in cached resources")
                content_data: str = cached_binary_resources[base_url + "/" + binary_url]
            else:
                logger.debug(f'Did not find Binary in cached resources, querying {base_url + "/" + binary_url}')
                binary_url_lookup: Response = session.get(f"{base_url}/{binary_url}", headers=query_headers)

                if binary_url_lookup.status_code != 200:
                    logger.error(f"The query responded with a status code of {binary_url_lookup.status_code}")
                    if binary_url_lookup.status_code == 403:
                        logger.error("The 403 code typically means your defined scope does not allow for retrieving this resource. Please check your scope to ensure it includes Binary.Read.")
                        if "WWW-Authenticate" in binary_url_lookup.headers:
                            logger.error(binary_url_lookup.headers["WWW-Authenticate"])
                    if binary_url_lookup.status_code == 400 and "json" in binary_url_lookup.headers["content-type"]:
                        logger.error(binary_url_lookup.json())
                try:
                    if binary_url_lookup.status_code == 200 and "json" in binary_url_lookup.headers["content-type"]:
                        content_data = binary_url_lookup.json()["data"]
                    elif binary_url_lookup.status_code == 200:
                        content_data = binary_url_lookup.text
                    else:
                        logger.warning("Skipping DocumentReference since Binary resource could not be retrieved")
                        return None
                except JSONDecodeError:
                    logger.warning("Skipping DocumentReference since Binary resource could not be retrieved")
                    logger.warning(f"Response code: {binary_url_lookup.status_code}")
                    logger.warning(f"Response text: {binary_url_lookup.content}")
                    logger.warning(f"Response headers: {binary_url_lookup.headers}")
                    return None
                cached_binary_resources[base_url + "/" + binary_url] = content_data

            resource["content"][i]["attachment"]["data"] = content_data
            del resource["content"][i]["attachment"]["url"]

    # Convert HTML to plain text
    html_contents = list(filter(lambda x: x["attachment"]["contentType"] == "text/html", resource["content"]))
    converted_htmls: list = []

    # This means the resource only has incompatible formats and needs to be removed from the returned Bundle
    if not html_contents:
        return None

    for content in html_contents:
        html_blurb: str = content["attachment"]["data"]
        text_maker = html2text.HTML2Text()
        text_maker.ignore_images = True
        text_blurb: str = text_maker.handle(html_blurb)
        text_blurb_bytes: bytes = text_blurb.encode("utf-8")
        base64_text: str = base64.b64encode(text_blurb_bytes).decode("utf-8")
        converted_htmls.append({"attachment": {"contentType": "text/plain", "data": base64_text}})

    resource["content"].extend(converted_htmls)

    plain_text_content: list[tuple[int, Any]] = [(idx, content) for idx, content in enumerate(resource["content"]) if content["attachment"]["contentType"] == "text/plain"]

    swap_contents: tuple = resource["content"][plain_text_content[0][0]], resource["content"][0]

    resource["content"][0], resource["content"][plain_text_content[0][0]] = swap_contents

    return resource


def expand_document_reference_content(entry_resource: dict) -> dict | None:
    """
    Expand content attachments of a DocumentReference resource into data fields.

    This function takes a Bunlde.entry where the entry.resource is of type DocumentReference and, for each content attachment that has a URL,
    it retrieves the data from the URL using an HTTP request and updates the content to include the data.

    Parameters:
    - entry_resource (dict): A Bundle.entry where resourceType is DocumentReference resource as a dictionary.

    Returns:
    - dict: The expanded DocumentReference resource with content data fields.

    Errors and Logging:
    - If the Binary retrieval fails (e.g., due to a non-200 status code), an error message is logged containing the status code and provides information about possible solutions for the error.
    - If a 403 status code is encountered, it suggests that the user's scope may be insufficient and provides guidance on checking the scope to ensure it includes 'Binary.Read'.
    - If the HTTP response contains 'WWW-Authenticate' headers, they are logged to provide additional diagnostic information.

    This function handles HTML attachments by converting them to plain text if necessary.
    """

    entry: dict[str, Any] = entry_resource.dict(exclude_none=True)  # type: ignore
    resource: dict[str, Any] = entry["resource"]

    if resource["resourceType"] == "OperationOutcome":
        handle_operation_outcomes(resource=resource)
        return resource

    for i, content in enumerate(resource["content"]):
        if "url" in content["attachment"]:
            binary_url: str = content["attachment"]["url"]
            if g_base_url + "/" + binary_url in cached_binary_resources:
                logger.debug("Found Binary in cached resources")
                content_data: str = cached_binary_resources[g_base_url + "/" + binary_url]
            else:
                logger.debug(f'Did not find Binary in cached resources, querying {g_base_url + "/" + binary_url}')
                binary_url_lookup: Response = g_session.get(f"{g_base_url}/{binary_url}", headers=g_query_headers)

                if binary_url_lookup.status_code != 200:
                    logger.error(f"The query responded with a status code of {binary_url_lookup.status_code}")
                    if binary_url_lookup.status_code == 403:
                        logger.error("The 403 code typically means your defined scope does not allow for retrieving this resource. Please check your scope to ensure it includes Binary.Read.")
                        if "WWW-Authenticate" in binary_url_lookup.headers:
                            logger.error(binary_url_lookup.headers["WWW-Authenticate"])
                    if binary_url_lookup.status_code == 400 and "json" in binary_url_lookup.headers["content-type"]:
                        logger.error(binary_url_lookup.json())
                try:
                    if binary_url_lookup.status_code == 200 and "json" in binary_url_lookup.headers["content-type"]:
                        content_data = binary_url_lookup.json()["data"]
                    elif binary_url_lookup.status_code == 200:
                        content_data = binary_url_lookup.text
                    else:
                        logger.warning("Skipping DocumentReference since Binary resource could not be retrieved")
                        return None
                except JSONDecodeError:
                    logger.warning("Skipping DocumentReference since Binary resource could not be retrieved")
                    logger.warning(f"Response code: {binary_url_lookup.status_code}")
                    logger.warning(f"Response text: {binary_url_lookup.content}")
                    logger.warning(f"Response headers: {binary_url_lookup.headers}")
                    return None
                cached_binary_resources[g_base_url + "/" + binary_url] = content_data

            resource["content"][i]["attachment"]["data"] = content_data
            del resource["content"][i]["attachment"]["url"]

    # Convert HTML to plain text
    html_contents = list(filter(lambda x: x["attachment"]["contentType"] == "text/html", resource["content"]))
    converted_htmls: list = []

    # This means the resource only has incompatible formats and needs to be removed from the returned Bundle
    if not html_contents:
        return None

    for content in html_contents:
        html_blurb: str = content["attachment"]["data"]
        text_maker = html2text.HTML2Text()
        text_maker.ignore_images = True
        text_blurb: str = text_maker.handle(html_blurb)
        text_blurb_bytes: bytes = text_blurb.encode("utf-8")
        base64_text: str = base64.b64encode(text_blurb_bytes).decode("utf-8")
        converted_htmls.append({"attachment": {"contentType": "text/plain", "data": base64_text}})

    resource["content"].extend(converted_htmls)

    plain_text_content: list[tuple[int, Any]] = [(idx, content) for idx, content in enumerate(resource["content"]) if content["attachment"]["contentType"] == "text/plain"]

    swap_contents: tuple = resource["content"][plain_text_content[0][0]], resource["content"][0]

    resource["content"][0], resource["content"][plain_text_content[0][0]] = swap_contents

    entry["resource"] = resource

    return entry


def expand_document_references_in_bundle(session: Session, input_bundle: Bundle, base_url: str, query_headers: dict = {}) -> Bundle:
    """
    Expand content attachments of DocumentReference entries within a Bundle.

    This function takes a FHIR Bundle containing DocumentReference resources and expands content attachments into data fields for each entry.

    Parameters:
    - input_bundle (Bundle): The input FHIR Bundle containing DocumentReference entries to be processed.
    - base_url (str): The base URL used for making HTTP requests to resolve content URLs.
    - query_headers (dict, optional): Additional headers for the HTTP requests (default: {}).

    Returns:
    - Bundle: A modified FHIR Bundle with expanded content data or the original input Bundle if an error occurs.
    """

    global cached_binary_resources, g_session, g_base_url, g_query_headers
    g_session, g_base_url, g_query_headers = session, base_url, query_headers

    returned_resources: list[BundleEntryType] = input_bundle.entry
    output_bundle: dict = deepcopy(input_bundle).dict(exclude_none=True)

    with ThreadPoolExecutor() as executor:
        future_to_entry: dict[Future[dict[str, Any] | None], BundleEntryType] = {executor.submit(expand_document_reference_content, entry): entry for entry in returned_resources}

        expanded_entries: list = []
        for future in as_completed(future_to_entry):
            entry: dict[str, Any] | None = future.result()
            if entry:
                expanded_entries.append(entry)

    expanded_entries_clean: list[dict[str, Any]] = [entry for entry in expanded_entries if entry]

    if len(cached_binary_resources.keys()) != 0:
        cached_binary_resources = {}

    output_bundle["entry"] = expanded_entries_clean
    output_bundle["total"] = len(expanded_entries_clean)
    return Bundle.parse_obj(output_bundle)
